package CmUtils::Objects::Element;
require 5.006;

use strict;
use warnings;
use Carp;

use CmUtils::Objects::Node;
use CmUtils::debug;

=head1  CmUtils::Objects::Element

TODO: This is a work in progress until this message is removed.

Defines an Element, which is an ordered collection of nodes. The node
references are stored an array in the same order as cmiss uses.

        	Xi1	Xi2	Xi3
node # 0	0	0	0
node # 1	1	0	0
node # 2	0	1	0
node # 3	1	1	0
node # 4	0	0	1
node # 5	1	0	1
node # 6	0	1	1
node # 7	1	1	1

=head1 Element structure

=over 4

=item NAME

Name of the Element

=item NODE

An ordered list of the CMISS::Nodes in the element

=item OCCURRENCE

A list of the number of each occurrence of the node. Corresponds
with NODE:
e.g. If NODE contains           1 2 3 2 5 2 7 8
then OCCURRENCE will contain    1 1 1 2 1 3 1 1

=item VERSION

A list of the number of each occurrence of the node. Corresponds
with NODE:
e.g. If NODE contains       1 2 3 2 5 2 7 8
then OCCURRENCE contains    1 1 1 2 1 3 1 1
and VERSION might be        1 5 1 4 1 9 1 1 .

The actual version numbers will be 1 if there is only one version
of the node or the real version number of the node if there is
more than one version. Also the version number may be greater than
1 even if none of the nodes in the element are shared.

=item COLLAPSED

Indicates the number of collapsed nodes in the element.

=back

=head1 VERSION

0.1 (14 May 2004) Glenn Ramsey

=head1 CHANGES

=head1 METHODS

=cut

BEGIN {
  our ( $VERSION, @ISA );

  $VERSION = 0.1;
  @ISA     = qw(CmUtils::debug);
}

=head2 B<new()>

Returns a new Element.

=cut

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {
    NAME        => undef,
    NODE       => [],
    OCCURRENCE => [],
    VERSION    => [],
    COLLAPSED   => 0
  };

  # options are
  #my $options = shift;

  # 	$options = {
  # 		VERSION => undef, # add versions to the node
  # 	}

  bless $self, $class;
  return $self;
}

=head2 B<copy()>

Returns an identical copy of an existing Element.

=cut

sub copy {
  my $self = shift;

  my $class = ref($self) || $self;

  my $new = $self->new();
  $new->name( $self->name() );
  $new->setNodes( $self->getNodes() );
  $new->setOccurrences( $self->getOccurrences() );
  $new->setVersions( $self->getVersions() );
  $new->{COLLAPSED} = $self->collapsed();
  return $new;
}

=head2 B<name([string/number])>

Optional argument is name for Element.  Returns the name of the Element.

=cut

sub name {
  my $self = shift;
  croak "usage: thing->name(name)" unless @_ <= 1;
  if (@_) { $self->{NAME} = shift }
  return $self->{NAME};
}

=head2 B<collapsed()>

Returns the number of collapsed nodes in the element.

=cut

sub collapsed {
  my $self = shift;
  croak "usage: thing->collapsed()" unless @_ <= 1;
  return $self->{COLLAPSED};
}

=head2 B<addNodes(Nodes)>

Adds nodes to the Element. The parameter is a list of Nodes and must
contain exactly 8 nodes.
 
=cut

sub addNodes {
  use CmUtils::File::Utils qw/strtonum/;

  my $self = shift;

  # "usage: thing->addNodes(values)"

  if ( $self->debug() ) {
    carp "Adding nodes to Element " . $self->name() . "\n";
  }
  # There is no reason to have this restriction here
  #croak "Only elements with 8 nodes are supported." if ( @_ != 8 );

  # count the number of occurrences of this node in the element
  # @_ should contain refs to CMISS::Node objects
  foreach my $node (@_) {

    if ( $self->debug() ) {
      carp "Node ".$node->name()."\n";
    }
    push @{ $self->{NODE} }, $node;

    my $occ = 1;
    # iterate backwards through the node array
    for ( my $j = scalar @{ $self->{NODE} } - 1 ;$j > 0 ; $j-- )
    {
      if ( $self->{NODE}[ $j - 1 ] == $node ) {

        #if(0==$occ){$occ++};
        $occ++;

        if ( $self->debug() ) {
          carp "Node "
            . $node->name()
            #. " :ref "
            #. $node
            . " occurrence "
            . $occ . "\n";
        }
      }
    }

    push @{ $self->{OCCURRENCE} }, $occ;
    push @{ $self->{VERSION} },    1;      #use version 1 by default

    # increment the count of collapsed nodes if necessary
    $self->{COLLAPSED}++ if ( $occ > 1 );

#    if ( $self->debug() ) {
#      carp "Element "
#        . $self->name()
#        . " added node "
#        . $node->name()
#        . " :ref "
#        . $node
#        . " occurrence "
#        . $#{ $self->{OCCURRENCE} } . "\n";
#    }
  }

  return $#{ $self->{NODE} };
}

=head2 B<clearValues()>

Clears the list of nodes in the Element.

=cut

sub clearNodes {
  my $self = shift;
  $self->{NODE}       = [];
  $self->{OCCURRENCE} = [];
  $self->{COLLAPSED}  = [];
}

=head2 B<setNodes()>

Sets the list of node names of the Element (clears, then adds the new values).

=cut

sub setNodes {
  my $self = shift;
  $self->clearNodes();
  return $self->addNodes(@_);
}

=head2 B<getNodes()>

Returns an array of the Nodes in the Element.

=cut

sub getNodes {
  my $self = shift;

  # "usage: thing->getNodes()"
  return @{ $self->{NODE} };
}

=head2 B<getNodeNames()>

Returns an array of the Node names in the Element.

=cut

sub getNodeNames {
  my $self = shift;

  # "usage: thing->getNodeNames()"
  my @nodenames;
  foreach my $node ( @{ $self->{NODE} } ) {
    push @nodenames, $node->name();
  }
  return @nodenames;
}

=head2 B<node(index,[value])>

Sets and/or return a single Node in a Element by index.

=cut

sub node {
  my $self = shift;
  croak "usage: thing->node(index)" unless @_ <= 1;
  my $number = shift;
  return $self->{NODE}[$number];
}

=head2 B<getOccurrences()>

Returns the occurrences array.

=cut

sub getOccurrences {
  my $self = shift;

  # "usage: thing->getNodes()"
  return @{ $self->{OCCURRENCE} };
}

=head2 B<occurrence(index,[value])>

Sets and/or returns a single item in the Element's occurrence array by index.

=cut

sub occurrence {
  my $self = shift;
  croak "usage: thing->occurrence(number,[value])" unless @_ <= 2;
  my $number = shift;
  if (@_) { $self->{OCCURRENCE}[$number] = shift }
  return $self->{OCCURRENCE}[$number];
}

=head2 B<nodeVersion(index,[value])>

Sets and/or returns a single item in the Element's version array by index.

=cut

sub nodeVersion {
  my $self = shift;
  croak "usage: thing->nodeVersion(number,[value])" unless @_ <= 2;
  my $number = shift;
  if (@_) { $self->{VERSION}[$number] = shift }
  return $self->{VERSION}[$number];
}

=head2 B<getNodeVersions(index,[value])>

Returns the Element's version array.

=cut

sub getNodeVersions {
  my $self = shift;

  # "usage: thing->getNodeVersions()"
  return @{ $self->{VERSION} };
}

=head2 B<setOptions(values)>

Set the options hash.

=cut

sub setOptions {
  my $self = shift;
  $self->clearOptions();
  return $self->addOption(@_);
}

=head2 B<clearOptions(index,[value])>

Clear the options hash.

=cut

sub clearOptions {
  my $self = shift;
  $self->{OPTIONS} = [];
}

=head2 B<getOptions(index,[value])>

Get the options hash.

=cut

sub getOptions {
  my $self = shift;
  return @{ $self->{OPTIONS} };
}

=head2 B<swapXi(Xidir1,Xidir2)>

Swaps the Xi directions in the element.
Takes 2 integer arguments that indicate the Xi directions to operate on.
e.g.
   1 2 swap Xi1 and Xi2
   1 3 swap Xi1 and Xi3
   2 1 same as 1 2
   1 1 reverse Xi1 direction
   2 2 reverse Xi1 direction

Warning: not tested very much. 

=cut

sub swapXi {
  # TODO: make this work for 2D and 1D elements
  my $self = shift;
  croak "usage: thing->swapXi( direction1, direction2)" unless @_ == 2;
  my $d1 = shift;
  my $d2 = shift;
  croak "direction indices must be 1,2 or 3"
    unless ( ( 1 <= $d1 ) && ( $d1 <= 3 ) && ( 1 <= $d2 ) && ( $d2 <= 3 ) );
  $d1--;
  $d2--;

  if ( $d1 != $d2 ) {
    my @first  = ( 2, 3, 5 );
    my @second = ( 7, 6, 4 );

    my $index1 = $first[$d1];
    my $index2 = $first[$d2];
    swap_nodes( $index1, $index2 );

    $index1 = $second[$d1];
    $index2 = $second[$d2];
    swap_nodes( $index1, $index2 );

  }
  else {

    # both indices the same - Xi dir reversal
    # Xi1 0<>1, 2<>3, 4<>5, 6<>7
    # Xi2 0<>2, 1<>3, 4<>6, 5<>7
    # Xi3 0<>4, 1<>5, 2<>5, 3<>7
    my $offset = 2**$d1;
    my $step   = 2**( $d1 + 1 );

    #print "$offset $step\n";
    #return;
    for ( my $i = 0 ; $i < 8 ; $i += $step ) {
      for ( my $j = 0 ; $j < $offset ; $j++, $i++ ) {

        #print "index1=".$i." index2=". ($i+$offset) ."\n";
        $self->swap_nodes( $i, $i + $offset );
      }
      $i = $i - $offset;
    }

  }

}

#used internally by swapXi
sub swap_nodes {
  my $self = shift;
  croak "usage: thing->swap_nodes(index1, index2)" unless @_ == 2;
  my $index1 = shift;
  my $index2 = shift;

  my $tmp;
  $tmp = @{ $self->{NODE} }[$index1];
  @{ $self->{NODE} }[$index1] = @{ $self->{NODE} }[$index2];
  @{ $self->{NODE} }[$index2] = $tmp;

}

=head2 B<list([FH])>

List the contents of this Element to the filehandle I<FH> (default to STDOUT).
Set the "verbose" flag ($s->verbose(1)) to see extended information.

=cut

sub list {
  my $self = shift;
  my $FH   = (@_) ? shift: \*STDOUT;

  my @nodes = $self->getNodes();
  printf $FH <<EOF, $self->name(), scalar @nodes;
   Element     %s
     Number of values       %d
     Index Name Ocrnce Vrsn
EOF
  my @occ  = $self->getOccurrences();
  my @vers = $self->getNodeVersions();
  for ( my $i = 0 ; $i < @nodes ; $i++ ) {
    printf $FH "     %4d     %s    %4d    %4d\n", $i, $nodes[$i]->name(),
      $occ[$i], $vers[$i];
  }
}

1;

__END__
